#include "main.h"
#include "always.h"

//by Shane Tolmie (c)2001 from http://www.workingtex.com/htpic

//*****
//METHOD 1
//writes a char or an int or long in decimal using putch_to_device(), skipping leading zeros
//problems: quite slow
//advantages: any number of digits, up to what a long can handle
void putdec(long c)
{
	static bit skip_zeros;
	long temp_c;
	long power_of_10;

	if (c==0)
	{
		putch_to_device('0');			    //otherwise it will skip all zeros altogether
	}
	else
	{
		skip_zeros=TRUE;							//skip initial zeros in number
		power_of_10=1000000000;				//does anything <=999,999,999
		do
		{
			power_of_10/=10;
			temp_c=c/power_of_10;
			//hundreds of thousands
			if (((temp_c)>0) || (skip_zeros==FALSE))
			{
				putch_to_device((unsigned char)((temp_c)+'0'));
				skip_zeros=FALSE;
			}
			c-=(temp_c)*power_of_10;
		}
		while (power_of_10!=1);
	}
}

//*****
//METHOD 2
//writes an *long* decimal to the default output device using putch_to_device()
//outputs decimals up to 7 characters (ie: 9,999,999)
//skips leading zeros
//advantages: fast
void putdec_long_to_device(unsigned long binary)
{
  union {
    struct {
      unsigned long low; //4 bytes
      unsigned long high;  //4 bytes
    } lng;
    unsigned char byte[8];
  } cdec;

  static bit encountered_only_zeros;
  signed char c;
  unsigned char d;

  /*  
	binary->bcd conversion method
	
  1. Add 3 to every BCD nibble that's >=5
  2. Shift left
  Repeat until binary portion is all 0000
  
  Example:  Convert $F9 to BCD (ie: 3 nibbles, containing 249) 
  
  0000 0000 0000 : 1111 1001
  0000 0000 0001 : 1111 0010  (shift left)
  0000 0000 0011 : 1110 0100  (shift left)
  0000 0000 0111 : 1100 1000  (shift left)
  0000 0000 1010 : 1100 1000  (add 3)
  0000 0001 0101 : 1001 0000  (shift left)
  0000 0001 1000 : 1001 0000  (add 3)
  0000 0011 0001 : 0010 0000  (shift left)
  0000 0110 0010 : 0100 0000  (shift left)
  0000 1001 0010 : 0100 0000  (add 3)
  0001 0010 0100 : 1000 0000  (shift left)
  0010 0100 1001 : 0000 0000  (shift left)
     2    4    9 = 249
  
  And the left now holds BCD 249.
  MUCH easier than messing with IDIV and FDIV!!!	
  
  This technique can be used for other mixed radix 
  conversions, the magic number 3 is half the difference 
  in the radix  (16-10)/2 = 3.

	For BCD->ascii, as in the 16-bit conversion below,
	any number >=8 must have 3 subtracted from it.
*/

  //at least it prints '0' if it is zero, and doesnt skip all leading zeros!
  if (binary==0) {
    putch_to_device('0');   
    return;
  }

  //implement the technique above
  cdec.lng.low=binary;
  cdec.lng.high=0;  //12W / 2l

  for (c=0;c<32;c++) {

    //shift hi:lo <<1
    d=cdec.byte[3];
    cdec.lng.low<<=1;
    cdec.lng.high<<=1;    
    if ( (d& 0x80)==0x80) cdec.byte[4]|=1;
      
    for (d=4;d<7;d++) {
      if ((cdec.byte[d]&0x0F)>=0x05) cdec.byte[d]+=0x3;
      if ((cdec.byte[d]&0xF0)>=0x50) cdec.byte[d]+=0x30;
    }
  } //65W

  //extract BCD and put to default output device
  
  encountered_only_zeros=TRUE;  //for skipping leading zeros

/*
  //METHOD 1
  //simple way of doing the loop below
  for (c=32;c>=0;c-=4) {
    d=(cdec.lng.high>>c)&0x0F;
    if (d>=8) d-=3;
    if (d!=0) encountered_only_zeros=FALSE;
    d+='0';	//convert to digits
    if (encountered_only_zeros==FALSE) {
      putch_to_device(d);    
    }
  } //49 words
*/

  //METHOD 2
  for (c=15;c>=8;c--) {
    d=cdec.byte[c>>1];
    if ((c&1)==0) {
      d&=0x0F;
    }
    else {
      d&=0xF0;
      d>>=4;
    }
    if (d>=8) d-=3;
    if (d!=0) encountered_only_zeros=FALSE;
    d+='0';	//convert to digits
    if (encountered_only_zeros==FALSE) {
      putch_to_device(d);   
    }
  } //37 words

} //114 words

